/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "pack.h"


static char ** bitmaps;
int leaves, full_blocks;


static void get_bitmap (struct super_block * s)
{
    int i, bmap_num;

    bmap_num = get_bmap_num (s);
    bitmaps = (char **)malloc (sizeof (char *) * bmap_num);
    if (!bitmaps)
	die ("malloc failed");
    for (i = 0; i < bmap_num; i ++)
	bitmaps[i] = SB_AP_BITMAP (s)[i]->b_data;
}


static int is_block_used (struct super_block * s, __u32 block)
{
    return test_bitmap_bit (bitmaps, block, s->s_blocksize);
}

static void mark_block_free (struct super_block * s, __u32 block)
{
    clear_bitmap_bit (bitmaps, block, s->s_blocksize);
}



// FIXME: sometimes we do not need offset even if it is not 0
static int pack_offset (struct packed_item * pi, struct item_head * ih, char * buf)
{
    if (get_offset (&ih->ih_key) > 0xffffffffULL) {
	pi->mask |= OFFSET64;
	*(__u64 *)buf = get_offset (&ih->ih_key);
	return sizeof (__u64);
    }

    pi->mask |= OFFSET32;
    *(__u32 *)buf = get_offset (&ih->ih_key);
    return sizeof (__u32);
}


// does not work for extents
static int pack_indirect (struct packed_item * pi, struct buffer_head * bh, 
			  struct item_head * ih, char * buf)
{
    int i;
    __u32 * indirect;
    char * p = buf;

    pi->mask |= (ITEM_BODY | INDIRECT_ITEM);
    if (ih_entry_count (ih)) {
	pi->mask |= ENTRY_COUNT;
	*(__u16 *)buf = ih_entry_count (ih);
	buf += sizeof (__u16);
    }

    indirect = (__u32 *)B_I_PITEM (bh, ih);
    for (i = 0; i < I_UNFM_NUM (ih); i ++) {
	if (buf + 2 * sizeof (__u32) - p > ih_item_len (ih)) {
	    // copy whole item
	    pi->mask |= WHOLE_INDIRECT;
	    memcpy (p + ((pi->mask & ENTRY_COUNT) ? sizeof (__u16) : 0), indirect, ih_item_len (ih));
	    return ih_item_len (ih) + ((pi->mask & ENTRY_COUNT) ? sizeof (__u16) : 0);
	}

	if (i && ((!indirect[i] && !indirect[i - 1]) || // hole continues
		  (indirect[i] == indirect[i - 1] + 1))) { // contiguous blocks
	    // update counter
	    *(__u32 *)buf = ((*(__u32 *)buf) + 1);
	} else {
	    // start new extent
	    if (i)
		// buf points to the counter of contiguous blocks (or zero
		// pointers)
		buf += sizeof (__u32);

	    *(__u32 *)buf = indirect[i];
	    buf += sizeof (__u32);
	    *(__u32 *)buf = 1;
	}
    }

    buf += sizeof (__u32);

    if ((buf - p - ((pi->mask & ENTRY_COUNT) ? sizeof (__u16) : 0)) % (2 * sizeof (__u32)))
	reiserfs_panic (0, "pack_indirect: %h of %lu packed into %d\n",
			ih, bh->b_blocknr, 
			(buf - p - ((pi->mask & ENTRY_COUNT) ? sizeof (__u16) : 0)));
    return buf - p;
}


// put whole item for now
static int pack_direntry (struct packed_item * pi, struct buffer_head * bh,
			  struct item_head * ih, char * buf)
{
    pi->mask |= (ENTRY_COUNT | ITEM_BODY | DIRENTRY_ITEM);
    *(__u16 *)buf = ih_entry_count (ih);
    buf += sizeof (__u16);
    memcpy (buf, B_I_PITEM (bh, ih), ih_item_len (ih));
    return sizeof (__u16) + ih_item_len (ih);
}


int pack_stat_data (struct packed_item * pi, struct buffer_head * bh,
		    struct item_head * ih, char * buf)
{
    int len = ih_item_len (ih);

    pi->mask |= ITEM_BODY;
    if (ih_entry_count (ih) != 0xffff) {
	pi->mask |= ENTRY_COUNT;
	*(__u16 *)buf = ih_entry_count (ih);
	buf += sizeof (__u16);
	len += sizeof (__u16);
    }
    memcpy (buf, B_I_PITEM (bh, ih), ih_item_len (ih));
    return len;
}


//
static void pack_leaf (struct buffer_head * bh)
{
    int i;
    char * buf, * start;
    struct item_head * ih;
    struct packed_item * pi;


    buf = malloc (bh->b_size);
    if (!buf)
	die ("malloc failed");

    start = buf;

    // start magic
    *(__u16 *)buf = LEAF_START_MAGIC;
    buf += sizeof (__u16);

    // block number
    *(__u32 *)buf = bh->b_blocknr;
    buf += sizeof (__u32);

    // item number
    *(__u16 *)buf = node_item_number (bh);
    buf += sizeof (__u16);

    ih = B_N_PITEM_HEAD (bh, 0);
    pi = (struct packed_item *)buf;
    for (i = 0; i < node_item_number (bh); i ++, ih ++) {
	buf = (char *)(pi + 1);
	pi->mask = 0;
	pi->item_len = ih_item_len (ih);

	// format
	if (get_offset (&ih->ih_key) && key_format (&ih->ih_key) != ih_key_format (ih))
	    reiserfs_warning ("key format does not match to item format %h\n", ih);
	if (ih_key_format (ih) == KEY_FORMAT_2)
	    pi->mask |= NEW_FORMAT;

	// k_dir_id
	if (!i || (i && ih->ih_key.k_dir_id != (ih - 1)->ih_key.k_dir_id)) {
	    pi->mask |= DIR_ID;
	    *(__u32 *)buf = ih->ih_key.k_dir_id;
	    buf += sizeof (__u32);
	}
	// k_object_id
	if (!i || (i && ih->ih_key.k_objectid != (ih - 1)->ih_key.k_objectid)) {
	    pi->mask |= OBJECT_ID;
	    *(__u32 *)buf = ih->ih_key.k_objectid;
	    buf += sizeof (__u32);
	}
	// k_offset
	if (get_offset (&ih->ih_key)) {
	    buf += pack_offset (pi, ih, buf);
	}

	// we do not need to store item type: 
	// when neither OFFSET32 nor OFFSET64 are set - this is stat data
	// when ENTRY_COUNT is set - this is directory
	// when ITEM_BODY is not set - this is a direct item
	if (is_direct_ih (ih)) {
	    ;
	} else if (is_indirect_ih (ih)) {
	    buf += pack_indirect (pi, bh, ih, buf);

	} else if (is_direntry_ih (ih)) {
	    buf += pack_direntry (pi, bh, ih, buf);

	} else if (is_stat_data_ih (ih)) {
	    buf += pack_stat_data (pi, bh, ih, buf);
	}

	// set size of the packed item
	pi->length = buf - (char *)pi;
	pi = (struct packed_item *)((char *)pi + pi->length);
    }
    *(__u16 *)buf = LEAF_END_MAGIC;
    buf += sizeof (__u16);
    
    fwrite (start, buf - start, 1, stdout);

    free (start);

    leaves ++;

    return;
}

extern int opt_pack_all;

static int get_total_block_number (struct super_block * s)
{
    int i, j;
    int retval = 0;
    int bmap_num;
    
    retval = 0;
    
    if (opt_pack_all)
	retval = get_block_count (s);
    else {
	bmap_num = get_bmap_num (s);
	for (i = 0; i < bmap_num; i ++) {
	    for (j = 0; j < s->s_blocksize * 8; j ++)
		if (i * s->s_blocksize * 8 + j < get_block_count (s) &&
		    test_bit (j, SB_AP_BITMAP (s)[i]->b_data))
		    retval ++;
	}
    }
    return retval;
}


static void pack_full_block (struct buffer_head * bh)
{
    __u16 magic;
    __u32 block;

    magic = FULL_BLOCK_START_MAGIC;
    fwrite (&magic, sizeof (__u16), 1, stdout);

    block = bh->b_blocknr;
    fwrite (&block, sizeof (__u32), 1, stdout);
    
    fwrite (bh->b_data, 4096, 1, stdout);

    full_blocks ++;
}


static void pack_super_block (struct super_block * s)
{
    pack_full_block (SB_BUFFER_WITH_SB (s));
    mark_block_free (s, SB_BUFFER_WITH_SB (s)->b_blocknr);
}


static void pack_bitmap (struct super_block * s)
{
    int i;
    int bmap_num;

    bmap_num = get_bmap_num (s);
    for (i = 0; i < bmap_num; i ++) {
	pack_full_block (SB_AP_BITMAP (s)[i]);
	mark_block_free (s, SB_AP_BITMAP (s)[i]->b_blocknr);
    }
}


static void pack_journal (struct super_block * s)
{
    int i;
    struct buffer_head * bh;

    if (is_prejournaled_reiserfs (s->s_rs))
	return;
    
    for (i = SB_JOURNAL_BLOCK (s);
	 i < SB_JOURNAL_BLOCK (s) + journal_size (s) + 1; i ++) {
	bh = bread (s->s_dev, i, s->s_blocksize);
	if (!bh)
	    continue;
	if (who_is_this (bh->b_data, bh->b_size) == THE_LEAF)
	    pack_leaf (bh);
	else
	    pack_full_block (bh);
	mark_block_free (s, bh->b_blocknr);
	brelse (bh);
    }
}


#define BLOCKS_PER_READ 8


// read blocks marked as used, pack leaves, mark leaves and internals as free
// which allows to have bitmap of unformatted nodes after this pass
static void pack_leaves (struct super_block * s)
{
    int i, j;
    struct buffer_head * bh;
    struct buffer_head tmp;
    int total_block_number, done;
    int block_count;

    // how many blocks have to be read. Needed to show progress
    total_block_number = get_total_block_number (s);
    done = 0;

    tmp.b_size = s->s_blocksize;
    block_count = get_block_count (s);

    for (i = 0; i < block_count / BLOCKS_PER_READ * BLOCKS_PER_READ; i += BLOCKS_PER_READ) {
	for (j = 0; j < BLOCKS_PER_READ; j ++)
	    if (is_block_used (s, i + j))
		break;
	if (j == BLOCKS_PER_READ)
	    // all blocks are free
	    continue;

	bh = bread (s->s_dev, i / BLOCKS_PER_READ, s->s_blocksize * BLOCKS_PER_READ);
	if (!bh)
	    continue;

	for (j = 0; j < BLOCKS_PER_READ; j ++) {
	    if (!is_block_used (s, i + j))
		continue;

	    tmp.b_blocknr = i + j;
	    tmp.b_data = bh->b_data + j * tmp.b_size;
	    if (who_is_this (tmp.b_data, tmp.b_size) == THE_LEAF) {
		pack_leaf (&tmp);
		mark_block_free (s, tmp.b_blocknr);
	    }
	    if (who_is_this (tmp.b_data, tmp.b_size) == THE_INTERNAL)
		mark_block_free (s, tmp.b_blocknr);

	    print_how_far (&done, total_block_number);
	}

	brelse (bh);
    }


    for (; i < block_count; i ++) {
	if (!is_block_used (s, i))
	    continue;
	bh = bread (s->s_dev, i, s->s_blocksize);
	if (!bh)
	    continue;
	if (who_is_this (bh->b_data, bh->b_size) == THE_LEAF) {
	    pack_leaf (&tmp);
	    mark_block_free (s, tmp.b_blocknr);
	}
	if (who_is_this (tmp.b_data, tmp.b_size) == THE_INTERNAL)
	    mark_block_free (s, tmp.b_blocknr);

	print_how_far (&done, total_block_number);

	brelse (bh);
    }
}


// bitmap is now bitmap of unformatted nodes
static void pack_unformatted_bitmap (struct super_block * s)
{
    __u32 magic;
    int i;
    __u16 v16;
    __u32 v32;

    magic = UNFORMATTED_BITMAP_START_MAGIC;
    fwrite (&magic, sizeof (__u16), 1, stdout);

    v16 = get_bmap_num (s);
    v32 = get_block_count (s);
    fwrite (&v16, sizeof (__u16), 1, stdout);
    fwrite (&v32, sizeof (__u32), 1, stdout);

    for (i = 0; i < v16; i ++)
	fwrite (SB_AP_BITMAP (s)[i]->b_data, s->s_blocksize, 1, stdout);

    magic = UNFORMATTED_BITMAP_END_MAGIC;
    fwrite (&magic, sizeof (__u16), 1, stdout);
}



// write reiserfs partition metadata to stdout in the following format:
//
// 32 bit reiserfs magic
// 16 bit blocksize
// blocks. Super block, bitmaps, journal blocks, leaves (packed)
// bitmaps of unformatted nodes (start magic, bitmap1, ..., end magic)
// 32 bit reiserfs magic
//
void pack_partition (struct super_block * s)
{
    __u32 magic32;
    __u16 blocksize;
    __u16 magic16;

    // get pointers to bitmaps
    get_bitmap (s);

    magic32 = REISERFS_SUPER_MAGIC;
    fwrite (&magic32, sizeof (__u32), 1, stdout);

    blocksize = s->s_blocksize;
    fwrite (&blocksize, sizeof (__u16), 1, stdout);

    // 1. 
    pack_super_block (s);

    // 2.
    pack_bitmap (s);

    // 3.
    pack_journal (s);

    // 4.
    pack_leaves (s);

    // 5.
    pack_unformatted_bitmap (s);
    
    magic16 = END_MAGIC;
    fwrite (&magic16, sizeof (__u16), 1, stdout);

    fprintf (stderr, "Packed\n\tleaves %d\n\tfull block %d\n", leaves, full_blocks);

    free (bitmaps);
}


void pack_one_block (struct super_block * s, struct buffer_head * bh)
{
    __u32 magic32;
    __u16 magic16;

    // reiserfs magic
    magic32 = REISERFS_SUPER_MAGIC;
    fwrite (&magic32, sizeof (__u32), 1, stdout);

    // blocksize
    fwrite (&(bh->b_size), sizeof (__u16), 1, stdout);
    
    if (who_is_this (bh->b_data, bh->b_size) == THE_LEAF)
	pack_leaf (bh);
    else
	pack_full_block (bh);

    // end magic
    magic16 = END_MAGIC;
    fwrite (&magic16, sizeof (__u16), 1, stdout);

}


#if 0
//
// this test program has two modes: 'pack file blocknr'
// and 'unpack file'
// in the first mode blocknr-th 4k block of the 'file' will be packed out to stdout
// the the second mode standart input will be converted to the reiserfs leaf on 'file'
//
static int do_unpack (char * file)
{
    char * buf;
    int fd;

    fd = open (file, O_RDONLY);
    if (fd == -1) {
	perror ("open failed");
	return 0;
    }
    
    buf = malloc (4096);
    if (!buf) {
	perror ("malloc failed");
	return 0;
    }

    fread (buf, 4096, 1, stdin);
    if (!feof (stdin)) {
	printf ("fread returned not eof\n");
	return 0;
    }

    unpack_leaf (buf, fd);

    free (buf);
    close (fd);
    return 0;
}

static int do_pack (char * file, int block)
{
    int fd;
    struct buffer_head * bh;
    char * buf;
    int len;

    fprintf (stderr, "dumping block %d of the \"%s\"\n", block, file);

    fd = open (file, O_RDONLY);
    if (fd == -1) {
	perror ("open failed");
	return 0;
    }
    
    bh = bread (fd, block, 4096);
    if (!bh) {
	fprintf (stderr, "bread failed\n");
	return 0;
    }

    if (who_is_this (bh->b_data, bh->b_size) != THE_LEAF) {
	fprintf (stderr, "block %d is not a leaf\n", block);
	return 0;
    }

    len = pack_leaf (bh, buf);
    fwrite (buf, len, 1, stdout);

    free (buf);
    close (fd);
    return 0;
}


int main (int argc, char ** argv)
{
    if (argc == 3 && !strcmp (argv[1], "unpack"))
	return do_unpack (argv[2]);

    if (argc == 4 && !strcmp (argv[1], "pack"))
	return do_pack (argv[2], atoi (argv[3]));

    fprintf (stderr, "Usage: \n\t%s pack filename block\nor\n"
	     "\t%s unpack filename\n", argv[0], argv[0]);
    return 0;
}

#endif
