/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */
  
#include "pack.h"


int leaves, full_blocks;



static void unpack_offset (int mask, struct item_head * ih)
{
    __u64 v;
    __u32 u;

    if (mask & OFFSET64) {
	if (ih_key_format (ih) != KEY_FORMAT_2)
	    die ("unpack_offset: key format is not set or wrong");
	fread64 (&v);
	set_offset (KEY_FORMAT_2, &ih->ih_key, v);
	return;
    }

    if (mask & OFFSET32) {
	fread32 (&u);
	set_offset (ih_key_format (ih), &ih->ih_key, u);
	return;
    }

    // offset is 0
    return;
}


// FIXME: it is not reliable
static int unpack_type (struct packed_item * pi, struct item_head * ih)
{
    if (!(pi->mask & ITEM_BODY)) {
	set_type (ih_key_format (ih), &ih->ih_key, TYPE_DIRECT);
	return 0;
    }

    if (!get_offset (&ih->ih_key) && !(pi->mask & INDIRECT_ITEM)) {
	set_type (ih_key_format (ih), &ih->ih_key, TYPE_STAT_DATA);
	return 0;
    }

    if (pi->mask & INDIRECT_ITEM) {
	set_type (ih_key_format (ih), &ih->ih_key, TYPE_INDIRECT);
	return 0;
    }
    
    if (pi->mask & DIRENTRY_ITEM) {
	set_type (ih_key_format (ih), &ih->ih_key, TYPE_DIRENTRY);
	return 0;
    }

    reiserfs_warning ("%h, mask 0%o\n", ih, pi->mask);
    fprintf (stderr, "unpack_type: can not define type of item\n");
    return 1;
    
}



static void unpack_direntry (struct packed_item * pi, struct buffer_head * bh,
			     int i, struct item_head * ih)
{
    __u16 v;

    fread16 (&v);
    set_entry_count (ih, v);

    set_ih_item_len (ih, pi->item_len);
    set_ih_location (ih, (i ? ih_location (ih - 1) : bh->b_size) - pi->item_len);
    fread (bh->b_data + ih_location (ih), pi->item_len, 1, stdin);
    return;
}


static void unpack_stat_data (struct packed_item * pi, struct buffer_head * bh,
			      int i, struct item_head * ih)
{
    __u16 v;

    if (pi->mask & ENTRY_COUNT) {
	fread16 (&v);
	set_entry_count (ih, v);
    } else
	set_entry_count (ih, 0xffff);

    set_ih_item_len (ih, pi->item_len);
    set_ih_location (ih, (i ? ih_location (ih - 1) : bh->b_size) - pi->item_len);

    fread (bh->b_data + ih_location (ih), pi->item_len, 1, stdin);
    return;
}


static void unpack_indirect (struct packed_item * pi, struct buffer_head * bh,
			     int i, struct item_head * ih)
{
    __u32 * indirect;
    int extents;
    int j, k, l;
    int indirect_len;
    __u16 v16;

    // calculate how many bytes the body of indirect item is stored in
    indirect_len = pi->length - sizeof (struct packed_item);

    if (pi->mask & DIR_ID)
	indirect_len -= sizeof (__u32);
    if (pi->mask & OBJECT_ID)
	indirect_len -= sizeof (__u32);
    if (pi->mask & OFFSET32)
	indirect_len -= sizeof (__u32);
    if (pi->mask & OFFSET64)
	indirect_len -= sizeof (__u64);

    if (pi->mask & ENTRY_COUNT) {
	fread16 (&v16);
	set_entry_count (ih, v16);
	indirect_len -= sizeof (__u16);
    } else {
	// ih_free_space is 0
    }
    set_ih_item_len (ih, pi->item_len);
    set_ih_location (ih, (i ? ih_location (ih - 1) : bh->b_size) - pi->item_len);

    indirect = (__u32 *)(bh->b_data + ih_location (ih));
    if (pi->mask & WHOLE_INDIRECT) {
	fread (indirect, pi->item_len, 1, stdin);
	return;
    }

    // number of extents
    extents = indirect_len / (2 * sizeof (__u32));
    if (indirect_len % (2 * sizeof (__u32)))
	reiserfs_panic (0, "unpack_indirect: block %d, indirect item %h packed into %d",
			bh->b_blocknr, ih, indirect_len);

    for (j = k = 0; j < extents; j ++) {
	__u32 start, num;

	fread32 (&start);
	fread32 (&num);
	for (l = 0; l < num; l ++) {
	    if (start)
		indirect [k++] = start + l;
	    else
		indirect [k++] = 0;
	}
    }
    
    if (k != I_UNFM_NUM (ih)) {
	reiserfs_warning ("k=%d, %h, extents %d\n", k, ih, extents);
	die ("unpack_indirect: could not unpack indirect item");
    }
    return;
}


// FIXME: we have no way to preserve symlinks
static void unpack_direct (struct packed_item * pi, struct buffer_head * bh,
			   int i, struct item_head * ih)
{
    set_entry_count (ih, 0xffff);
    set_ih_item_len (ih, pi->item_len);
    set_ih_location (ih, (i ? ih_location (ih - 1) : bh->b_size) - pi->item_len);
    memset (bh->b_data + ih_location (ih), 'a', pi->item_len);
    return;
}


static void unpack_leaf (int dev)
{
    struct buffer_head * bh;
    int block, item_number;
    struct packed_item pi;
    struct item_head * ih;
    int i;
    __u16 v16;
    __u32 v32;
    
    // block number
    fread32 (&v32);

    //printf ("Unpacking leaf %d\n", v32);

    // item number
    fread16 (&v16);
    item_number = v16;
    
    bh = getblk (dev, v32, 4096);
    if (!bh)
	die ("unpack_leaf: getblk failed");

    set_node_item_number (bh, item_number);
    set_node_level (bh, DISK_LEAF_NODE_LEVEL);
    set_node_free_space (bh, bh->b_size - BLKH_SIZE);


    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < item_number; i ++, ih ++) {
	fread (&pi, sizeof (struct packed_item), 1, stdin);

	if (bh->b_blocknr == 13231)
	    fprintf (stderr, "#%d: dir_id %d, obj_id %d, off32 %d, off64 %d, "
		     "entrycount/freespace %d, new format %d, item body %d"
		     "whole indirect %d, indirect item %d, direntry %d\n", i,
		     pi.mask & DIR_ID, pi.mask & OBJECT_ID, pi.mask & OFFSET32, pi.mask & OFFSET64,
		     pi.mask & ENTRY_COUNT, pi.mask & NEW_FORMAT, pi.mask & ITEM_BODY,
		     pi.mask & WHOLE_INDIRECT, pi.mask & INDIRECT_ITEM,
		     pi.mask & DIRENTRY_ITEM);
	
	// dir_id
	if (pi.mask & DIR_ID) {
	    fread32 (&v32);
	    ih->ih_key.k_dir_id = v32;
	} else {
	    if (!i)
		die ("unpack_leaf: dir_id is not set");
	    ih->ih_key.k_dir_id = (ih - 1)->ih_key.k_dir_id;
	}

	// objectid
	if (pi.mask & OBJECT_ID) {
	    fread32 (&v32);
	    ih->ih_key.k_objectid = v32;
	} else {
	    if (!i)
		die ("unpack_leaf: object_id is not set");
	    ih->ih_key.k_objectid = (ih - 1)->ih_key.k_objectid;
	}

	// we need to set item format before offset unpacking
	set_ih_key_format (ih, (pi.mask & NEW_FORMAT) ? KEY_FORMAT_2 : KEY_FORMAT_1);

	// offset
	unpack_offset (pi.mask, ih);

	// type
	if (unpack_type (&pi, ih)) {
	    printf ("block %d, could not unpack item %d\n", bh->b_blocknr, i);
	}

	// item itself
	if (is_direct_ih (ih)) {
	    unpack_direct (&pi, bh, i, ih);
	} else if (is_indirect_ih (ih)) {
	    unpack_indirect (&pi, bh, i, ih);
	} else if (is_direntry_ih (ih)) {
	    unpack_direntry (&pi, bh, i, ih);
	} else if (is_stat_data_ih (ih)) {
	    unpack_stat_data (&pi, bh, i, ih);
	}
	set_node_free_space (bh, node_free_space (bh) - (IH_SIZE + ih_item_len (ih)));
    }

    fread16 (&v16);
    if (v16 != LEAF_END_MAGIC)
	die ("unpack_leaf: wrong end signature found");

    mark_buffer_uptodate (bh, 1);
    mark_buffer_dirty (bh, 1);

    // leaf is unpacked into bh

    //    print_block (bh, 3, -1, -1);
    //
    bwrite (bh);
    brelse (bh);
    leaves ++;
}


static void unpack_full_block (int dev, int blocksize)
{
    __u16 v16;
    __u32 block;
    struct buffer_head * bh;

    fread32 (&block);
    bh = getblk (dev, block, blocksize);
    if (!bh)
	die ("unpack_full_block: getblk failed");

    fread (bh->b_data, bh->b_size, 1, stdin);
    mark_buffer_uptodate (bh, 1);
    mark_buffer_dirty (bh, 1);
    bwrite (bh);
    brelse (bh);
    full_blocks ++;
}


int do_not_overwrite_block_to_be_unformatted = 1;

static void unpack_unformatted_bitmap (int dev, int blocksize)
{
    __u16 bmap_num;
    __u32 block_count;
    int i, j, k;
    char * buf;
    int unformatted, done;
    struct buffer_head * bh;

    fread16 (&bmap_num);
    fread32 (&block_count);

    buf = malloc (blocksize);
    if (!buf)
	die ("unpack_unformatted_bitmap: malloc failed");

    fprintf (stderr, "Resetting unformatted..\n");
    for (i = 0; i < bmap_num; i ++) {
	fread (buf, blocksize, 1, stdin);
	if (do_not_overwrite_block_to_be_unformatted)
	    // -n option was specified, so user is sure, that target
	    // partition does not have blocks looking like leaves
	    continue;

	// get number of unformatted nodes pointed by this bitmap
	unformatted = 0;
	for (j = 0; j < blocksize * 8 && i * blocksize * 8 + j < block_count; j ++)
	    if (test_bit (j, buf))
		unformatted ++;

	done = 0;
	if (unformatted) {
	    fprintf (stderr, "bitmap #%d (out of %d, %d blocks) - ", i, bmap_num, unformatted);
	    for (j = 0; j < blocksize * 8 && i * blocksize * 8 + j < block_count; j ++) {
		if (test_bit (j, buf)) {
		    bh = getblk (dev, i * blocksize * 8 + j, blocksize);
		    if (!bh)
			die ("unpack_unformatted_bitmap: getblk failed");
		    mark_buffer_uptodate (bh, 1);
		    mark_buffer_dirty (bh, 1);
		    bwrite (bh);
		    brelse (bh);
		    print_how_far (&done, unformatted);
		}
	    }
	    printf ("\n");
	}
    }

    fread16 (&bmap_num);
    if (bmap_num != UNFORMATTED_BITMAP_END_MAGIC)
	die ("unpack_unformatted_bitmap: no ending magic found");
}


// read packed reiserfs partition metadata from stdin
void unpack_partition (int dev)
{
    __u32 magic32;
    __u16 magic16;
    __u16 blocksize;

    fread32 (&magic32);
    if (magic32 != REISERFS_SUPER_MAGIC)
	die ("unpack_partition: reiserfs magic number not found");

    fread16 (&blocksize);

    while (!feof (stdin)) {
	fread16 (&magic16);
	switch (magic16) {
	case LEAF_START_MAGIC:
	    unpack_leaf (dev);
	    break;

	case FULL_BLOCK_START_MAGIC:
	    unpack_full_block (dev, blocksize);
	    break;

	case UNFORMATTED_BITMAP_START_MAGIC:
	    unpack_unformatted_bitmap (dev, blocksize);
	    break;

	case END_MAGIC:
	    break;

	default:
	    die ("unpack_partition: bad magic found");
	}
    }

    fprintf (stderr, "Unpacked\n\tleaves %d\n\tfull block %d\n", leaves, full_blocks);
}


int main (int argc, char ** argv)
{
    int fd;

    if (argc < 2) {
	fprintf (stderr, "Usage: %s filename\n", argv[0]);
	return 0;
    }
    
    fd = open (argv[1], O_RDWR);
    if (fd == -1) {
	perror ("open failed");
	return 0;
    }

    unpack_partition (fd);

    close (fd);
    return 0;
}
