/*
 *  Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "include.h"


/* read super block and bitmaps. fixme: only 4k blocks, pre-journaled format
   is refused */
reiserfs_filsys_t * reiserfs_open (char * filename, int flags)
{
    reiserfs_filsys * fs;
    struct buffer_head * bh;
    struct reiserfs_super_block * rs;
    int fd, i;
    unsigned long block;
    
    fd = open (filename, flags);
    if (fd == -1) {
	reiserfs_warning ("reiserfs_open: could not open %d due to %s\n",
			  filename, strerror (errno));
	return 0;
    }

    /* reiserfs super block is either in 16-th or in 2-nd 4k block of the
       device */
    for (i = 16; i > 0; i -= 14) {
	bh = bread (fd, i, 4096);
	if (!bh) {
	    reiserfs_warning ("reiserfs_open: bread failed reading block %d\n", i);
	} else {
	    rs = (struct reiserfs_super_block *)bh->b_data;
	    
	    if (is_reiser2fs_magic_string (rs) || is_reiserfs_magic_string (rs))
		goto found;

	    /* reiserfs signature is not found at the i-th 4k block */
	    brelse (bh);
	}
    }

    reiserfs_warning ("reiserfs_open: neither new nor old reiserfs format "
		      "found on %s\n", filename);
    return 0;

 found:

    /* fixme: we could make some check to make sure that super block looks
       correctly */
    fs = getmem (sizeof (reiserfs_filsys));
    fs->s_dev = fd;
    fs->s_blocksize = rs_blocksize (rs);
    fs->s_hash_function = hashes [rs_hash (rs)];
    SB_BUFFER_WITH_SB (fs) = bh;
    s->u.reiserfs_sb.s_rs = rs;
    s->s_flags = flags; /* O_RDONLY or O_RDWR */

    /* read bitmaps */
    SB_AP_BITMAP (fs) = getmem (sizeof (void *) * s_bmap_nr (rs));
    for (i = 0, block = bh->b_blocknr + 1; i < rs_bmap_nr (rs); i ++) {
	SB_AP_BITMAP (fs)[i] = bread (fd, block, fs->s_blocksize);
	if (!SB_AP_BITMAP (fs)[i]) {
	    reiserfs_warning ("reiserfs_open: bread failed reading bitmap #%d (%lu)\n", i, block);
	    SB_AP_BITMAP (fs)[i] = getblk (fd, block, fs->s_blocksize);
	    memset (SB_AP_BITMAP (fs)[i]->b_data, 0xff, fs->s_blocksize);
	    set_bit (BH_Uptodate, &SB_AP_BITMAP (fs)[i]->b_state);
	}
	block = (bh->b_blocknr == 16 ? ((i + 1) * fs->s_blocksize * 8) : (block + 1));
    }

    return 0;
    
}


/* flush all changes made on a filesystem */
void reiserfs_flush (reiserfs_filsys * fs)
{
    flush_buffers (fs->s_dev);
}


/* free all memory involved into manipulating with filesystem */
void reiserfs_free (reiserfs_filsys * fs)
{
    int i;

    /* release bitmaps */
    for (i = 0; i < SB_BMAP_NR (fs); i ++)
	brelse (SB_AP_BITMAP (fs) [i]);
    freemem (SB_AP_BITMAP (fs));

    /* release super block and memory used by filesystem handler */
    SB_BUFFER_WITH_SB (fs) = bh;
    freemem (fs);

    free_buffers ();
}


void reiserfs_close (reiserfs_filsys * fs)
{
    reiserfs_flush (fs);
    reiserfs_free (fs);
}




